@@ -1,4 +1,4 @@
-/* copyright (c) 2022 - 2023 grunfink / MIT license */
+/* copyright (c) 2022 - 2023 grunfink et al. / MIT license */
 
 #ifndef _XS_H
 
@@ -65,12 +65,15 @@ xs_str *xs_str_new(const char *str);
 xs_str *xs_str_wrap_i(const char *prefix, xs_str *str, const char *suffix);
 #define xs_str_prepend_i(str, prefix) xs_str_wrap_i(prefix, str, NULL)
 #define xs_str_cat(str, suffix) xs_str_wrap_i(NULL, str, suffix)
-xs_str *xs_replace_i(xs_str *str, const char *sfrom, const char *sto);
-#define xs_replace(str, sfrom, sto) xs_replace_i(xs_dup(str), sfrom, sto)
+xs_str *xs_replace_in(xs_str *str, const char *sfrom, const char *sto, int times);
+#define xs_replace_i(str, sfrom, sto) xs_replace_in(str, sfrom, sto, XS_ALL)
+#define xs_replace(str, sfrom, sto) xs_replace_in(xs_dup(str), sfrom, sto, XS_ALL)
+#define xs_replace_n(str, sfrom, sto, times) xs_replace_in(xs_dup(str), sfrom, sto, times)
 xs_str *xs_fmt(const char *fmt, ...);
 int xs_str_in(const char *haystack, const char *needle);
-int xs_startswith(const char *str, const char *prefix);
-int xs_endswith(const char *str, const char *postfix);
+int xs_starts_and_ends(const char *prefix, const char *str, const char *suffix);
+#define xs_startswith(str, prefix) xs_starts_and_ends(prefix, str, NULL)
+#define xs_endswith(str, suffix) xs_starts_and_ends(NULL, str, suffix)
 xs_str *xs_crop_i(xs_str *str, int start, int end);
 xs_str *xs_strip_chars_i(xs_str *str, const char *chars);
 #define xs_strip_i(str) xs_strip_chars_i(str, " \r\n\t\v\f")
@@ -81,7 +84,7 @@ xs_list *xs_list_append_m(xs_list *list, const char *mem, int dsz);
 #define xs_list_append(list, data) xs_list_append_m(list, data, xs_size(data))
 int xs_list_iter(xs_list **list, xs_val **value);
 int xs_list_len(const xs_list *list);
-char *xs_list_get(const xs_list *list, int num);
+xs_val *xs_list_get(const xs_list *list, int num);
 xs_list *xs_list_del(xs_list *list, int num);
 xs_list *xs_list_insert(xs_list *list, int num, const xs_val *data);
 xs_list *xs_list_insert_sorted(xs_list *list, const char *str);
@@ -93,6 +96,7 @@ int xs_list_in(const xs_list *list, const xs_val *val);
 xs_str *xs_join(const xs_list *list, const char *sep);
 xs_list *xs_split_n(const char *str, const char *sep, int times);
 #define xs_split(str, sep) xs_split_n(str, sep, XS_ALL)
+xs_list *xs_list_cat(xs_list *l1, const xs_list *l2);
 
 xs_dict *xs_dict_new(void);
 xs_dict *xs_dict_append_m(xs_dict *dict, const xs_str *key, const xs_val *mem, int dsz);
@@ -100,7 +104,7 @@ xs_dict *xs_dict_append_m(xs_dict *dict, const xs_str *key, const xs_val *mem, i
 xs_dict *xs_dict_prepend_m(xs_dict *dict, const xs_str *key, const xs_val *mem, int dsz);
 #define xs_dict_prepend(dict, key, data) xs_dict_prepend_m(dict, key, data, xs_size(data))
 int xs_dict_iter(xs_dict **dict, xs_str **key, xs_val **value);
-xs_dict *xs_dict_get(const xs_dict *dict, const xs_str *key);
+xs_val *xs_dict_get(const xs_dict *dict, const xs_str *key);
 xs_dict *xs_dict_del(xs_dict *dict, const xs_str *key);
 xs_dict *xs_dict_set(xs_dict *dict, const xs_str *key, const xs_val *data);
 
@@ -115,6 +119,11 @@ void xs_data_get(const xs_data *value, void *data);
 
 void *xs_memmem(const char *haystack, int h_size, const char *needle, int n_size);
 
+xs_str *xs_hex_enc(const xs_val *data, int size);
+xs_val *xs_hex_dec(const xs_str *hex, int *size);
+int xs_is_hex(const char *str);
+
+unsigned int xs_hash_func(const char *data, int size);
 
 #ifdef XS_ASSERT
 #include <assert.h>
@@ -125,9 +134,27 @@ void *xs_memmem(const char *haystack, int h_size, const char *needle, int n_size
 #define XS_ASSERT_TYPE_NULL(v, t) (void)(0)
 #endif
 
+extern xs_val xs_stock_null[];
+extern xs_val xs_stock_true[];
+extern xs_val xs_stock_false[];
+extern xs_val xs_stock_0[];
+extern xs_val xs_stock_1[];
+extern xs_val xs_stock_list[];
+extern xs_val xs_stock_dict[];
+
+#define xs_return(v) xs_val *__r = v; v = NULL; return __r
+
 
 #ifdef XS_IMPLEMENTATION
 
+xs_val xs_stock_null[]  = { XSTYPE_NULL };
+xs_val xs_stock_true[]  = { XSTYPE_TRUE };
+xs_val xs_stock_false[] = { XSTYPE_FALSE };
+xs_val xs_stock_0[]     = { XSTYPE_NUMBER, '0', '\0' };
+xs_val xs_stock_1[]     = { XSTYPE_NUMBER, '1', '\0' };
+xs_val xs_stock_list[]  = { XSTYPE_LIST, 0, 0, 5, XSTYPE_EOM };
+xs_val xs_stock_dict[]  = { XSTYPE_DICT, 0, 0, 5, XSTYPE_EOM };
+
 void *_xs_realloc(void *ptr, size_t size, const char *file, int line, const char *func)
 {
     d_char *ndata = realloc(ptr, size);
@@ -409,30 +436,37 @@ xs_str *xs_str_wrap_i(const char *prefix, xs_str *str, const char *suffix)
         str = xs_insert_m(str, 0, prefix, strlen(prefix));
 
     if (suffix)
-        str = xs_insert_m(str, xs_size(str) - 1, suffix, xs_size(suffix));
+        str = xs_insert_m(str, strlen(str), suffix, strlen(suffix));
 
     return str;
 }
 
 
-xs_str *xs_replace_i(xs_str *str, const char *sfrom, const char *sto)
+xs_str *xs_replace_in(xs_str *str, const char *sfrom, const char *sto, int times)
 /* replaces inline all sfrom with sto */
 {
     XS_ASSERT_TYPE(str, XSTYPE_STRING);
 
     int sfsz = strlen(sfrom);
     int stsz = strlen(sto);
+    int diff = stsz - sfsz;
     char *ss;
     int offset = 0;
 
-    while ((ss = strstr(str + offset, sfrom)) != NULL) {
+    while (times > 0 && (ss = strstr(str + offset, sfrom)) != NULL) {
         int n_offset = ss - str;
 
-        str = xs_collapse(str, n_offset, sfsz);
-        str = xs_expand(str, n_offset, stsz);
+        if (diff < 0)
+            str = xs_collapse(str, n_offset, -diff);
+        else
+        if (diff > 0)
+            str = xs_expand(str, n_offset, diff);
+
         memcpy(str + n_offset, sto, stsz);
 
         offset = n_offset + stsz;
+
+        times--;
     }
 
     return str;
@@ -475,20 +509,23 @@ int xs_str_in(const char *haystack, const char *needle)
 }
 
 
-int xs_startswith(const char *str, const char *prefix)
-/* returns true if str starts with prefix */
+int xs_starts_and_ends(const char *prefix, const char *str, const char *suffix)
+/* returns true if str starts with prefix and ends with suffix */
 {
-    return !!(xs_str_in(str, prefix) == 0);
-}
+    int sz = strlen(str);
+    int psz = prefix ? strlen(prefix) : 0;
+    int ssz = suffix ? strlen(suffix) : 0;
 
+    if (sz < psz || sz < ssz)
+        return 0;
 
-int xs_endswith(const char *str, const char *postfix)
-/* returns true if str ends with postfix */
-{
-    int ssz = strlen(str);
-    int psz = strlen(postfix);
+    if (prefix && memcmp(str, prefix, psz) != 0)
+        return 0;
 
-    return !!(ssz >= psz && memcmp(postfix, str + ssz - psz, psz) == 0);
+    if (suffix && memcmp(str + sz - ssz, suffix, ssz) != 0)
+        return 0;
+
+    return 1;
 }
 
 
@@ -554,15 +591,10 @@ xs_str *xs_tolower_i(xs_str *str)
 xs_list *xs_list_new(void)
 /* creates a new list */
 {
-    xs_list *list;
-
-    list = xs_realloc(NULL, _xs_blk_size(5));
-    list[0] = XSTYPE_LIST;
-    list[4] = XSTYPE_EOM;
-
-    _xs_put_24b(list + 1, 5);
-
-    return list;
+    return memcpy(
+        xs_realloc(NULL, _xs_blk_size(sizeof(xs_stock_list))),
+        xs_stock_list, sizeof(xs_stock_list)
+    );
 }
 
 
@@ -571,10 +603,10 @@ xs_list *_xs_list_write_litem(xs_list *list, int offset, const char *mem, int ds
 {
     XS_ASSERT_TYPE(list, XSTYPE_LIST);
 
-    char c = XSTYPE_LITEM;
+    list = xs_expand(list, offset, dsz + 1);
 
-    list = xs_insert_m(list, offset,     &c,  1);
-    list = xs_insert_m(list, offset + 1, mem, dsz);
+    list[offset] = XSTYPE_LITEM;
+    memcpy(list + offset + 1, mem, dsz);
 
     return list;
 }
@@ -844,20 +876,26 @@ xs_list *xs_split_n(const char *str, const char *sep, int times)
 }
 
 
+xs_list *xs_list_cat(xs_list *l1, const xs_list *l2)
+/* concatenates list l2 to l1 */
+{
+    XS_ASSERT_TYPE(l1, XSTYPE_LIST);
+    XS_ASSERT_TYPE(l2, XSTYPE_LIST);
+
+    /* inserts at the end of l1 the content of l2 (skipping header and footer) */
+    return xs_insert_m(l1, xs_size(l1) - 1, l2 + 4, xs_size(l2) - 5);
+}
+
+
 /** dicts **/
 
 xs_dict *xs_dict_new(void)
 /* creates a new dict */
 {
-    xs_dict *dict;
-
-    dict = xs_realloc(NULL, _xs_blk_size(5));
-    dict[0] = XSTYPE_DICT;
-    dict[4] = XSTYPE_EOM;
-
-    _xs_put_24b(dict + 1, 5);
-
-    return dict;
+    return memcpy(
+        xs_realloc(NULL, _xs_blk_size(sizeof(xs_stock_dict))),
+        xs_stock_dict, sizeof(xs_stock_dict)
+    );
 }
 
 
@@ -980,7 +1018,7 @@ xs_dict *xs_dict_set(xs_dict *dict, const xs_str *key, const xs_val *data)
     dict = xs_dict_del(dict, key);
 
     /* add the data */
-    dict = xs_dict_prepend(dict, key, data);
+    dict = xs_dict_append(dict, key, data);
 
     return dict;
 }
@@ -1055,7 +1093,7 @@ const char *xs_number_str(const xs_number *v)
 }
 
 
-/* raw data blocks */
+/** raw data blocks **/
 
 xs_data *xs_data_new(const void *data, int size)
 /* returns a new raw data value */
@@ -1063,12 +1101,12 @@ xs_data *xs_data_new(const void *data, int size)
     xs_data *v;
 
     /* add the overhead (data type + 24bit size) */
-    size += 4;
+    int total_size = size + 4;
 
-    v = xs_realloc(NULL, _xs_blk_size(size));
+    v = xs_realloc(NULL, _xs_blk_size(total_size));
     v[0] = XSTYPE_DATA;
 
-    _xs_put_24b(v + 1, size);
+    _xs_put_24b(v + 1, total_size);
 
     memcpy(&v[4], data, size);
 
@@ -1109,6 +1147,87 @@ void *xs_memmem(const char *haystack, int h_size, const char *needle, int n_size
 }
 
 
+/** hex **/
+
+xs_str *xs_hex_enc(const xs_val *data, int size)
+/* returns an hexdump of data */
+{
+    xs_str *s;
+    char *p;
+    int n;
+
+    p = s = xs_realloc(NULL, _xs_blk_size(size * 2 + 1));
+
+    for (n = 0; n < size; n++) {
+        snprintf(p, 3, "%02x", (unsigned char)data[n]);
+        p += 2;
+    }
+
+    *p = '\0';
+
+    return s;
+}
+
+
+xs_val *xs_hex_dec(const xs_str *hex, int *size)
+/* decodes an hexdump into data */
+{
+    int sz = strlen(hex);
+    xs_val *s = NULL;
+    char *p;
+    int n;
+
+    if (sz % 2)
+        return NULL;
+
+    p = s = xs_realloc(NULL, _xs_blk_size(sz / 2 + 1));
+
+    for (n = 0; n < sz; n += 2) {
+        int i;
+        if (sscanf(&hex[n], "%02x", &i) == 0) {
+            /* decoding error */
+            return xs_free(s);
+        }
+        else
+            *p = i;
+
+        p++;
+    }
+
+    *p = '\0';
+    *size = sz / 2;
+
+    return s;
+}
+
+
+int xs_is_hex(const char *str)
+/* returns 1 if str is an hex string */
+{
+    while (*str) {
+        if (strchr("0123456789abcdefABCDEF", *str++) == NULL)
+            return 0;
+    }
+
+    return 1;
+}
+
+
+unsigned int xs_hash_func(const char *data, int size)
+/* a general purpose hashing function */
+{
+    unsigned int hash = 0x666;
+    int n;
+
+    for (n = 0; n < size; n++) {
+        hash ^= data[n];
+        hash *= 111111111;
+    }
+
+    return hash ^ hash >> 16;
+}
+
+
 #endif /* XS_IMPLEMENTATION */
 
 #endif /* _XS_H */
